#include <iostream>
#include <chrono>

int main() {
  std::cout << "running matrix transpose example" << std::endl;

  // config
  uint64_t l_matrix_size = 16384;
  uint64_t l_n_repetitions = 10;

  // sizes of the matrices
  std::size_t l_size_bytes  = l_matrix_size * l_matrix_size;
              l_size_bytes *= sizeof(double);
  double l_size_gib = l_size_bytes / (1024.0 * 1024.0 * 1024.0);

  // time measurement vars
  std::chrono::steady_clock::time_point l_tp0, l_tp1;
  std::chrono::duration< double > l_dur;

  // share config
  std::cout << "allocating uint64_t input and output matrix:\n"
            << "  dimensions:            " << l_matrix_size << "x" << l_matrix_size << "\n"
            << "  size per matrix (B):   " << l_size_bytes << "\n"
            << "  size per matrix (GiB): " << l_size_gib
            << std::endl;

  // allocate memory
  uint64_t * l_a = (uint64_t *) aligned_alloc( 128,
                                               l_size_bytes );

  uint64_t * l_b = (uint64_t *) aligned_alloc( 128,
                                               l_size_bytes );

  // init matrices
  std::cout << "initializing matrices" << std::endl;
  for( uint64_t l_co = 0; l_co < l_matrix_size; l_co++ ) {
    for( uint64_t l_ro = 0; l_ro < l_matrix_size; l_ro++ ) {
      uint64_t l_id = l_co * l_matrix_size + l_ro;
      l_a[l_id] = l_id;
      l_b[l_id] = 0;
    }
  }

  /*
   * vanilla transpose
   */
  std::cout << "benchmarking vanilla transpose" << std::endl;
  l_tp0 = std::chrono::steady_clock::now();
  for( uint64_t l_re = 0; l_re < l_n_repetitions; l_re++ ) {
    for( uint64_t l_co = 0; l_co < l_matrix_size; l_co++ ) {
      for( uint64_t l_ro = 0; l_ro < l_matrix_size; l_ro++ ) {
        uint64_t l_id_a = l_co * l_matrix_size + l_ro;
        uint64_t l_id_b = l_ro * l_matrix_size + l_co;

        l_b[l_id_b] = l_a[l_id_a];
      }
    }
  }
  l_tp1 = std::chrono::steady_clock::now();

  // checking results
  for( uint64_t l_co = 0; l_co < l_matrix_size; l_co++ ) {
    for( uint64_t l_ro = 0; l_ro < l_matrix_size; l_ro++ ) {
      uint64_t l_id  = l_co * l_matrix_size + l_ro;
      uint64_t l_val = l_ro * l_matrix_size + l_co;

      if( l_b[l_id] != l_val ) {
        std::cerr << "  transpose failed:\n"
                  << "    column:   " << l_co << "\n"
                  << "    row:      " << l_ro << "\n"
                  << "    value:    " << l_b[l_id] << "\n"
                  << "    expected: " << l_val
                  << std::endl;

        std::cerr << "  aborting.." << std::endl;
        return EXIT_FAILURE;
        }
    }
  }

  // report performance
  l_dur = std::chrono::duration_cast< std::chrono::duration< double> >( l_tp1 - l_tp0 );
  std::cout << "  performance:\n"
            << "    time:      " << l_dur.count() << "\n"
            << "    bandwidth: " << l_size_gib*2*l_n_repetitions / l_dur.count() << " GB/s"
            << std::endl;

  /*
   * TODO: Add your optimized implementation here.
   */

  std::cout << "freeing memory" << std::endl;
  free(l_a);
  free(l_b);

  std::cout << "finished, exiting now" << std::endl;

  return EXIT_SUCCESS;
}
